@@ -18,7 +18,7 @@
 -- License     :  BSD-style (see the file LICENSE)
 --
 -- Maintainer  : "Samuel Gélineau" <gelisam@gmail.com>,
---               "Oleg Grenrus" <oleg.grenrus@iki.fi>,
+--               "Luc Tielen" <luc.tielen@gmail.com>,
 --               "Ryan Scott" <ryan.gl.scott@gmail.com>
 -- Stability   :  experimental
 -- Portability :  non-portable
@@ -26,73 +26,85 @@
 ----------------------------------------------------------------------------
 module Data.Functor.Foldable
   (
-  -- * Base functors for fixed points
+  -- * Base functors
     Base
   , ListF(..)
-  -- * Fixed points
-  , Fix(..), unfix
-  , Mu(..), hoistMu
-  , Nu(..), hoistNu
-  -- * Folding
-  , Recursive(..)
-  -- ** Combinators
-  , gapo
-  , gcata
-  , zygo
-  , gzygo
+  -- * Type classes
+  , Recursive(project)
+  , Corecursive(embed)
+  -- * Folding functions
+  -- $foldingFunctions
+  , fold
+  , cata
+  , cataA
+  , para
   , histo
-  , ghisto
+  , zygo
+  -- * Unfolding functions
+  , unfold
+  , ana
+  , apo
   , futu
-  , gfutu
+  -- * Combining unfolds and folds
+  , refold
+  , hylo
   , chrono
+  -- * Changing representation
+  , refix
+  , hoist
+  , transverse
+  , cotransverse
+  -- * Advanced usage
+  -- ** Mendler-style recursion-schemes
+  , mcata
+  , mpara
+  , mhisto
+  , mzygo
+  , mana
+  , mapo
+  , mfutu
+  -- ** Fokkinga's recursion-schemes
+  , prepro
+  , postpro
+  -- ** Elgot (co)algebras
+  , elgot
+  , coelgot
+  -- ** Generalized recursion-schemes
+  , gfold
+  , gcata
+  , gpara
+  , ghisto
+  , gzygo
+  , gunfold
+  , gana
+  , gapo
+  , gfutu
+  , grefold
+  , ghylo
   , gchrono
-  -- ** Distributive laws
+  , gprepro
+  , gpostpro
   , distCata
   , distPara
   , distParaT
-  , distZygo
-  , distZygoT
   , distHisto
   , distGHisto
-  , distFutu
-  , distGFutu
-  -- * Unfolding
-  , Corecursive(..)
-  -- ** Combinators
-  , gana
-  -- ** Distributive laws
+  , distZygo
+  , distZygoT
   , distAna
   , distApo
   , distGApo
   , distGApoT
-  -- * Refolding
-  , hylo
-  , ghylo
-  -- ** Changing representation
-  , hoist
-  , refix
-  -- * Common names
-  , fold, gfold
-  , unfold, gunfold
-  , refold, grefold
-  -- * Mendler-style
-  , mcata
-  , mhisto
-  -- * Elgot (co)algebras
-  , elgot
-  , coelgot
-  -- * Zygohistomorphic prepromorphisms
+  , distFutu
+  , distGFutu
+  -- ** Zygohistomorphic prepromorphisms
   , zygoHistoPrepro
-  -- * Effectful combinators
-  , cataA
-  , transverse
-  , cotransverse
   ) where
 
 import Control.Applicative
 import Control.Comonad
 import Control.Comonad.Trans.Class
-import Control.Comonad.Trans.Env
+import Control.Comonad.Trans.Env (EnvT(..))
 import qualified Control.Comonad.Cofree as Cofree
 import Control.Comonad.Cofree (Cofree(..))
 import           Control.Comonad.Trans.Cofree (CofreeF, CofreeT(..))
@@ -105,58 +117,179 @@ import           Control.Monad.Trans.Free (FreeF, FreeT(..))
 import qualified Control.Monad.Trans.Free as CMTF
 import Data.Functor.Identity
 import Control.Arrow
-import Data.Function (on)
-import Data.Functor.Classes
 import Data.Functor.Compose (Compose(..))
 import Data.List.NonEmpty(NonEmpty((:|)), nonEmpty, toList)
-import Text.Read
-import Text.Show
+import Data.Tree (Tree (..))
 #ifdef __GLASGOW_HASKELL__
-import Data.Data hiding (gunfold)
-#if HAS_POLY_TYPEABLE
-#else
-import qualified Data.Data as Data
-#endif
 #if HAS_GENERIC
 import GHC.Generics (Generic (..), M1 (..), V1, U1, K1 (..), (:+:) (..), (:*:) (..))
 #endif
-#if HAS_GENERIC1
-import GHC.Generics (Generic1)
-#endif
 #endif
 import Numeric.Natural
-import Data.Monoid (Monoid (..))
 import Prelude
 
-import qualified Data.Foldable as F
-import qualified Data.Traversable as T
-
-import qualified Data.Bifunctor as Bi
-import qualified Data.Bifoldable as Bi
-import qualified Data.Bitraversable as Bi
-
 import           Data.Functor.Base hiding (head, tail)
 import qualified Data.Functor.Base as NEF (NonEmptyF(..))
 
--- $setup
--- >>> :set -XDeriveFunctor
--- >>> import Control.Monad (void)
--- >>> import Data.Char (toUpper)
+import Data.Fix (Fix (..), unFix, Mu (..), Nu (..))
 
+-- $setup
+-- >>> :set -XDeriveFunctor -XScopedTypeVariables -XLambdaCase -XGADTs -XFlexibleContexts
+-- >>> import Control.Applicative (Const (..), Applicative (..))
+-- >>> import Control.Comonad
+-- >>> import Control.Comonad.Cofree (Cofree(..))
+-- >>> import Control.Monad (void)
+-- >>> import Control.Monad.Trans.Reader (Reader, ask, local, runReader)
+-- >>> import Data.Char (toUpper)
+-- >>> import Data.Fix (Fix (..))
+-- >>> import Data.Foldable (traverse_)
+-- >>> import Data.List (intercalate, partition)
+-- >>> import Data.List.NonEmpty (NonEmpty (..))
+-- >>> import Data.Maybe (maybeToList)
+-- >>> import Data.Tree (Tree (..), drawTree)
+-- >>> import Numeric.Natural
+--
+-- >>> import Data.Functor.Base
+--
+-- >>> let showTree = putStrLn . go where go (Node x xs) = if null xs then x else "(" ++ unwords (x : map go xs) ++ ")"
+--
+-- >>> let myTree = Node 0 [Node 1 [], Node 2 [], Node 3 [Node 31 [Node 311 [Node 3111 [], Node 3112 []]]]]
+
+-- $foldingFunctions
+-- Folding functions allow you to reduce a recursive structure down to a value. The value can be a simple type such as 'Int' or 'String', or it can also be a recursive structure. Each of the functions below will be accompanied by an example which folds the following @Tree Int@ down to some 'String'.
+--
+-- >>> putStr $ drawTree $ fmap show myTree
+-- 0
+-- |
+-- +- 1
+-- |
+-- +- 2
+-- |
+-- `- 3
+--    |
+--    `- 31
+--       |
+--       `- 311
+--          |
+--          +- 3111
+--          |
+--          `- 3112
+
+-- | Obtain the base functor for a recursive datatype.
+--
+-- The core idea of this library is that instead of writing recursive functions
+-- on a recursive datatype, we prefer to write non-recursive functions on a
+-- related, non-recursive datatype we call the "base functor".
+--
+-- For example, @[a]@ is a recursive type, and its corresponding base functor is
+-- @'ListF' a@:
+--
+-- @
+-- data 'ListF' a b = 'Nil' | 'Cons' a b
+-- type instance 'Base' [a] = 'ListF' a
+-- @
+--
+-- The relationship between those two types is that if we replace @b@ with
+-- @'ListF' a@, we obtain a type which is isomorphic to @[a]@.
+--
 type family Base t :: * -> *
 
+-- | A recursive datatype which can be unrolled one recursion layer at a time.
+--
+-- For example, a value of type @[a]@ can be unrolled into a @'ListF' a [a]@.
+-- Ifthat unrolled value is a 'Cons', it contains another @[a]@ which can be
+-- unrolled as well, and so on.
+--
+-- Typically, 'Recursive' types also have a 'Corecursive' instance, in which
+-- case 'project' and 'embed' are inverses.
 class Functor (Base t) => Recursive t where
+  -- | Unroll a single recursion layer.
+  --
+  -- >>> project [1,2,3]
+  -- Cons 1 [2,3]
   project :: t -> Base t t
 #ifdef HAS_GENERIC
   default project :: (Generic t, Generic (Base t t), GCoerce (Rep t) (Rep (Base t t))) => t -> Base t t
   project = to . gcoerce . from
 #endif
 
-  cata :: (Base t a -> a) -- ^ a (Base t)-algebra
-       -> t               -- ^ fixed point
-       -> a               -- ^ result
+  -- | An alias for 'fold'.
+  --
+  -- 'fold' is by far the most common recursion-scheme, because working one layer at a time is the most common strategy for writing a recursive function. But there are also other, rarer strategies. Researchers have given names to the most common strategies, and their name for 'fold' is "catamorphism". They also give its @Base t a -> a@ argument a special name, "(@Base t@)-algebra". More generally, a function of the form @f a -> a@ is called an "f-algebra".
+  --
+  -- The names might seem intimidating at first, but using the standard nomenclature has benefits. If you program with others, it can be useful to have a shared vocabulary to refer to those recursion patterns. For example, you can discuss which type of recursion is the most appropriate for the problem at hand. Names can also help to structure your thoughts while writing recursive functions.
+  --
+  -- The rest of this module lists a few of the other recursion-schemes which are common enough to have a name. In this section, we restrict our attention to those which fold a recursive structure down to a value. In the examples all functions will be of type @Tree Int -> String@.
+  cata :: (Base t a -> a) -> t -> a
   cata f = c where c = f . fmap c . project
 
+  -- | A variant of 'cata' in which recursive positions also include the
+  -- original sub-tree, in addition to the result of folding that sub-tree.
+  --
+  -- For our running example, let's add a number to each node indicating how
+  -- many children are below it. To do so, we will need to count those nodes
+  -- from the original sub-tree.
+  --
+  -- >>> :{
+  -- let pprint4 :: Tree Int -> String
+  --     pprint4 = flip runReader 0 . para go
+  --       where
+  --         go :: TreeF Int (Tree Int, Reader Int String)
+  --            -> Reader Int String
+  --         go (NodeF i trss) = do
+  --           -- trss :: [(Tree Int, Reader Int String)]
+  --           -- ts   :: [Tree Int]
+  --           -- rss  :: [Reader Int String]
+  --           -- ss   :: [String]
+  --           let (ts, rss) = unzip trss
+  --           let count = sum $ fmap length ts
+  --           ss <- local (+ 2) $ sequence rss
+  --           indent <- ask
+  --           let s = replicate indent ' '
+  --                ++ "* " ++ show i
+  --                ++ " (" ++ show count ++ ")"
+  --           pure $ intercalate "\n" (s : ss)
+  -- :}
+  --
+  -- >>> putStrLn $ pprint4 myTree
+  -- * 0 (7)
+  --   * 1 (0)
+  --   * 2 (0)
+  --   * 3 (4)
+  --     * 31 (3)
+  --       * 311 (2)
+  --         * 3111 (0)
+  --         * 3112 (0)
+  --
+  -- One common use for 'para' is to construct a new tree which reuses most of
+  -- the sub-trees from the original. In the following example, we insert a new
+  -- node under the leftmost leaf. This requires allocating new nodes along a
+  -- path from the root to that leaf, while keeping every other sub-tree
+  -- untouched.
+  --
+  -- >>> :{
+  -- let insertLeftmost :: Int -> Tree Int -> Tree Int
+  --     insertLeftmost new = para go
+  --       where
+  --         go :: TreeF Int (Tree Int, Tree Int)
+  --            -> Tree Int
+  --         go (NodeF i []) = Node i [Node new []]
+  --         go (NodeF i ((_orig, recur) : tts))
+  --             -- tts :: [(Tree Int, Tree Int)]
+  --           = let (origs, _recurs) = unzip tts
+  --             in Node i (recur : origs)
+  -- :}
+  --
+  -- >>> putStrLn $ pprint4 $ insertLeftmost 999 myTree
+  -- * 0 (8)
+  --   * 1 (1)
+  --     * 999 (0)
+  --   * 2 (0)
+  --   * 3 (4)
+  --     * 31 (3)
+  --       * 311 (2)
+  --         * 3111 (0)
+  --         * 3112 (0)
   para :: (Base t (t, a) -> a) -> t -> a
   para t = p where p x = t . fmap ((,) <*> p) $ project x
 
@@ -188,13 +321,27 @@ distPara = distZygo embed
 distParaT :: (Corecursive t, Comonad w) => (forall b. Base t (w b) -> w (Base t b)) -> Base t (EnvT t w a) -> EnvT t w (Base t a)
 distParaT t = distZygoT embed t
 
+-- | A recursive datatype which can be rolled up one recursion layer at a time.
+--
+-- For example, a value of type @'ListF' a [a]@ can be rolled up into a @[a]@.
+-- This @[a]@ can then be used in a 'Cons' to construct another @'ListF' a [a]@,
+-- which can be rolled up as well, and so on.
+--
+-- Typically, 'Corecursive' types also have a 'Recursive' instance, in which
+-- case 'embed' and 'project' are inverses.
 class Functor (Base t) => Corecursive t where
+
+  -- | Roll up a single recursion layer.
+  --
+  -- >>> embed (Cons 1 [2,3])
+  -- [1,2,3]
   embed :: Base t t -> t
 #ifdef HAS_GENERIC
   default embed :: (Generic t, Generic (Base t t), GCoerce (Rep (Base t t)) (Rep t)) => Base t t -> t
   embed = to . gcoerce . from
 #endif
 
+  -- | An alias for 'unfold'.
   ana
     :: (a -> Base t a) -- ^ a (Base t)-coalgebra
     -> a               -- ^ seed
@@ -223,105 +370,77 @@ class Functor (Base t) => Corecursive t where
     -> t
   gpostpro k e g = a . return where a = embed . fmap (hoist e . a . join) . k . liftM g
 
+-- | An alias for 'refold'.
 hylo :: Functor f => (f b -> b) -> (a -> f a) -> a -> b
 hylo f g = h where h = f . fmap h . g
 
+-- | Folds a recursive type down to a value, one layer at a time.
+--
+-- >>> :{
+-- let mySum :: [Int] -> Int
+--     mySum = fold $ \case
+--       Nil -> 0
+--       Cons x sumXs -> x + sumXs
+-- :}
+--
+-- >>> mySum [10,11,12]
+-- 33
+--
+-- In our running example, one layer consists of an 'Int' and a list of recursive positions. In @Tree Int@, those recursive positions contain sub-trees of type @Tree Int@. Since we are working one layer at a time, the @Base t a -> a@ function is not given a @Tree Int@, but a @TreeF Int String@. That is, each recursive position contains the 'String' resulting from recursively folding the corresponding sub-tree.
+--
+-- >>> :{
+-- let pprint1 :: Tree Int -> String
+--     pprint1 = fold $ \case
+--       NodeF i [] -> show i
+--       NodeF i ss -> show i ++ ": [" ++ intercalate ", " ss ++ "]"
+-- :}
+--
+-- >>> putStrLn $ pprint1 myTree
+-- 0: [1, 2, 3: [31: [311: [3111, 3112]]]]
+--
+-- More generally, the 't' argument is the recursive value, the 'a' is the final result, and the @Base t a -> a@ function explains how to reduce a single layer full of recursive results down to a result.
 fold :: Recursive t => (Base t a -> a) -> t -> a
 fold = cata
 
+-- | A generalization of 'unfoldr'. The starting seed is expanded into a base
+-- functor whose recursive positions contain more seeds, which are themselves
+-- expanded, and so on.
+--
+-- >>> :{
+-- >>> let ourEnumFromTo :: Int -> Int -> [Int]
+-- >>>     ourEnumFromTo lo hi = ana go lo where
+-- >>>         go i = if i > hi then Nil else Cons i (i + 1)
+-- >>> :}
+--
+-- >>> ourEnumFromTo 1 4
+-- [1,2,3,4]
 unfold :: Corecursive t => (a -> Base t a) -> a -> t
 unfold = ana
 
+-- | An optimized version of @fold f . unfold g@.
+--
+-- Useful when your recursion structure is shaped like a particular recursive
+-- datatype, but you're neither consuming nor producing that recursive datatype.
+-- For example, the recursion structure of quick sort is a binary tree, but its
+-- input and output is a list, not a binary tree.
+--
+-- >>> data BinTreeF a b = Tip | Branch b a b deriving (Functor)
+--
+-- >>> :{
+-- >>> let quicksort :: Ord a => [a] -> [a]
+-- >>>     quicksort = refold merge split where
+-- >>>         split []     = Tip
+-- >>>         split (x:xs) = let (l, r) = partition (<x) xs in Branch l x r
+-- >>>
+-- >>>         merge Tip            = []
+-- >>>         merge (Branch l x r) = l ++ [x] ++ r
+-- >>> :}
+--
+-- >>> quicksort [1,5,2,8,4,9,8]
+-- [1,2,4,5,8,8,9]
 refold :: Functor f => (f b -> b) -> (a -> f a) -> a -> b
 refold = hylo
 
--- | Base functor of @[]@.
-data ListF a b = Nil | Cons a b
-  deriving (Eq,Ord,Show,Read,Typeable
-#if HAS_GENERIC
-          , Generic
-#endif
-#if HAS_GENERIC1
-          , Generic1
-#endif
-          )
-
-#ifdef LIFTED_FUNCTOR_CLASSES
-instance Eq2 ListF where
-  liftEq2 _ _ Nil        Nil          = True
-  liftEq2 f g (Cons a b) (Cons a' b') = f a a' && g b b'
-  liftEq2 _ _ _          _            = False
-
-instance Eq a => Eq1 (ListF a) where
-  liftEq = liftEq2 (==)
-
-instance Ord2 ListF where
-  liftCompare2 _ _ Nil        Nil          = EQ
-  liftCompare2 _ _ Nil        _            = LT
-  liftCompare2 _ _ _          Nil          = GT
-  liftCompare2 f g (Cons a b) (Cons a' b') = f a a' `mappend` g b b'
-
-instance Ord a => Ord1 (ListF a) where
-  liftCompare = liftCompare2 compare
-
-instance Show a => Show1 (ListF a) where
-  liftShowsPrec = liftShowsPrec2 showsPrec showList
-
-instance Show2 ListF where
-  liftShowsPrec2 _  _ _  _ _ Nil        = showString "Nil"
-  liftShowsPrec2 sa _ sb _ d (Cons a b) = showParen (d > 10)
-    $ showString "Cons "
-    . sa 11 a
-    . showString " "
-    . sb 11 b
-
-instance Read2 ListF where
-  liftReadsPrec2 ra _ rb _ d = readParen (d > 10) $ \s -> nil s ++ cons s
-    where
-      nil s0 = do
-        ("Nil", s1) <- lex s0
-        return (Nil, s1)
-      cons s0 = do
-        ("Cons", s1) <- lex s0
-        (a,      s2) <- ra 11 s1
-        (b,      s3) <- rb 11 s2
-        return (Cons a b, s3)
-
-instance Read a => Read1 (ListF a) where
-  liftReadsPrec = liftReadsPrec2 readsPrec readList
-
-#else
-instance Eq a   => Eq1   (ListF a) where eq1        = (==)
-instance Ord a  => Ord1  (ListF a) where compare1   = compare
-instance Show a => Show1 (ListF a) where showsPrec1 = showsPrec
-instance Read a => Read1 (ListF a) where readsPrec1 = readsPrec
-#endif
-
--- These instances cannot be auto-derived on with GHC <= 7.6
-instance Functor (ListF a) where
-  fmap _ Nil        = Nil
-  fmap f (Cons a b) = Cons a (f b)
-
-instance F.Foldable (ListF a) where
-  foldMap _ Nil        = Data.Monoid.mempty
-  foldMap f (Cons _ b) = f b
-
-instance T.Traversable (ListF a) where
-  traverse _ Nil        = pure Nil
-  traverse f (Cons a b) = Cons a <$> f b
-
-instance Bi.Bifunctor ListF where
-  bimap _ _ Nil        = Nil
-  bimap f g (Cons a b) = Cons (f a) (g b)
-
-instance Bi.Bifoldable ListF where
-  bifoldMap _ _ Nil        = mempty
-  bifoldMap f g (Cons a b) = mappend (f a) (g b)
-
-instance Bi.Bitraversable ListF where
-  bitraverse _ _ Nil        = pure Nil
-  bitraverse f g (Cons a b) = Cons <$> f a <*> g b
-
 type instance Base [a] = ListF a
 instance Recursive [a] where
   project (x:xs) = Cons x xs
@@ -345,6 +464,12 @@ instance Recursive (NonEmpty a) where
 instance Corecursive (NonEmpty a) where
   embed = (:|) <$> NEF.head <*> (maybe [] toList <$> NEF.tail)
 
+type instance Base (Tree a) = TreeF a
+instance Recursive (Tree a) where
+  project (Node x xs) = NodeF x xs
+instance Corecursive (Tree a) where
+  embed (NodeF x xs) = Node x xs
+
 type instance Base Natural = Maybe
 instance Recursive Natural where
   project 0 = Nothing
@@ -445,8 +570,9 @@ grefold, ghylo
   -> (a -> f (m a))
   -> a
   -> b
-ghylo w m f g = extract . h . return where
-  h = fmap f . w . fmap (duplicate . h . join) . m . liftM g
+ghylo w m f g = f . fmap (hylo alg coalg) . g where
+  coalg = fmap join . m . liftM g
+  alg   = fmap f . w . fmap duplicate
 grefold w m f g a = ghylo w m f g a
 
 futu :: Corecursive t => (a -> Base t (Free (Base t) a)) -> a -> t
@@ -469,82 +595,29 @@ distGFutu k = d where
 -- Fix
 -------------------------------------------------------------------------------
 
-newtype Fix f = Fix (f (Fix f))
-
-unfix :: Fix f -> f (Fix f)
-unfix (Fix f) = f
-
-instance Eq1 f => Eq (Fix f) where
-  Fix a == Fix b = eq1 a b
-
-instance Ord1 f => Ord (Fix f) where
-  compare (Fix a) (Fix b) = compare1 a b
-
-instance Show1 f => Show (Fix f) where
-  showsPrec d (Fix a) =
-    showParen (d >= 11)
-      $ showString "Fix "
-      . showsPrec1 11 a
-
-instance Read1 f => Read (Fix f) where
-  readPrec = parens $ prec 10 $ do
-    Ident "Fix" <- lexP
-    Fix <$> step (readS_to_Prec readsPrec1)
-
-#ifdef __GLASGOW_HASKELL__
-#if HAS_POLY_TYPEABLE
-deriving instance Typeable Fix
-deriving instance (Typeable f, Data (f (Fix f))) => Data (Fix f)
-#else
-instance Typeable1 f => Typeable (Fix f) where
-   typeOf t = mkTyConApp fixTyCon [typeOf1 (undefined `asArgsTypeOf` t)]
-     where asArgsTypeOf :: f a -> Fix f -> f a
-           asArgsTypeOf = const
-
-fixTyCon :: TyCon
-#if MIN_VERSION_base(4,4,0)
-fixTyCon = mkTyCon3 "recursion-schemes" "Data.Functor.Foldable" "Fix"
-#else
-fixTyCon = mkTyCon "Data.Functor.Foldable.Fix"
-#endif
-{-# NOINLINE fixTyCon #-}
-
-instance (Typeable1 f, Data (f (Fix f))) => Data (Fix f) where
-  gfoldl f z (Fix a) = z Fix `f` a
-  toConstr _ = fixConstr
-  gunfold k z c = case constrIndex c of
-    1 -> k (z (Fix))
-    _ -> error "gunfold"
-  dataTypeOf _ = fixDataType
-
-fixConstr :: Constr
-fixConstr = mkConstr fixDataType "Fix" [] Prefix
-
-fixDataType :: DataType
-fixDataType = mkDataType "Data.Functor.Foldable.Fix" [fixConstr]
-#endif
-#endif
-
 type instance Base (Fix f) = f
 instance Functor f => Recursive (Fix f) where
   project (Fix a) = a
 instance Functor f => Corecursive (Fix f) where
   embed = Fix
 
+-- | Convert from one recursive type to another.
+--
+-- >>> showTree $ hoist (\(NonEmptyF h t) -> NodeF [h] (maybeToList t)) ( 'a' :| "bcd")
+-- (a (b (c d)))
+--
 hoist :: (Recursive s, Corecursive t)
       => (forall a. Base s a -> Base t a) -> s -> t
 hoist n = cata (embed . n)
 
+-- | Convert from one recursive representation to another.
+--
+-- >>> refix ["foo", "bar"] :: Fix (ListF String)
+-- Fix (Cons "foo" (Fix (Cons "bar" (Fix Nil))))
+--
 refix :: (Recursive s, Corecursive t, Base s ~ Base t) => s -> t
 refix = cata embed
 
-toFix :: Recursive t => t -> Fix (Base t)
-toFix = refix
-
-fromFix :: Corecursive t => Fix (Base t) -> t
-fromFix = refix
-
-
 -------------------------------------------------------------------------------
 -- Lambek
 -------------------------------------------------------------------------------
@@ -557,7 +630,6 @@ lambek = cata (fmap embed)
 colambek :: (Recursive t, Corecursive t) => (Base t t -> t)
 colambek = ana (fmap project)
 
-newtype Mu f = Mu (forall a. (f a -> a) -> a)
 type instance Base (Mu f) = f
 instance Functor f => Recursive (Mu f) where
   project = lambek
@@ -565,27 +637,12 @@ instance Functor f => Recursive (Mu f) where
 instance Functor f => Corecursive (Mu f) where
   embed m = Mu (\f -> f (fmap (fold f) m))
 
-instance (Functor f, Eq1 f) => Eq (Mu f) where
-  (==) = (==) `on` toFix
-
-instance (Functor f, Ord1 f) => Ord (Mu f) where
-  compare = compare `on` toFix
-
-instance (Functor f, Show1 f) => Show (Mu f) where
-  showsPrec d f = showParen (d > 10) $
-    showString "fromFix " . showsPrec 11 (toFix f)
-
-#ifdef __GLASGOW_HASKELL__
-instance (Functor f, Read1 f) => Read (Mu f) where
-  readPrec = parens $ prec 10 $ do
-    Ident "fromFix" <- lexP
-    fromFix <$> step readPrec
-#endif
-
--- | A specialized, faster version of 'hoist' for 'Mu'.
-hoistMu :: (forall a. f a -> g a) -> Mu f -> Mu g
-hoistMu n (Mu mk) = Mu $ \roll -> mk (roll . n)
-
+type instance Base (Nu f) = f
+instance Functor f => Corecursive (Nu f) where
+  embed = colambek
+  ana = Nu
+instance Functor f => Recursive (Nu f) where
+  project (Nu f a) = Nu f <$> f a
 
 -- | Church encoded free monads are Recursive/Corecursive, in the same way that
 -- 'Mu' is.
@@ -599,37 +656,7 @@ instance Functor f => Corecursive (CMFC.F f a) where
   embed (CMTF.Pure a)  = CMFC.F $ \p _ -> p a
   embed (CMTF.Free fr) = CMFC.F $ \p f -> f $ fmap (cmfcCata p f) fr
 
-
-data Nu f where Nu :: (a -> f a) -> a -> Nu f
-type instance Base (Nu f) = f
-instance Functor f => Corecursive (Nu f) where
-  embed = colambek
-  ana = Nu
-instance Functor f => Recursive (Nu f) where
-  project (Nu f a) = Nu f <$> f a
-
-instance (Functor f, Eq1 f) => Eq (Nu f) where
-  (==) = (==) `on` toFix
-
-instance (Functor f, Ord1 f) => Ord (Nu f) where
-  compare = compare `on` toFix
-
-instance (Functor f, Show1 f) => Show (Nu f) where
-  showsPrec d f = showParen (d > 10) $
-    showString "fromFix " . showsPrec 11 (toFix f)
-
-#ifdef __GLASGOW_HASKELL__
-instance (Functor f, Read1 f) => Read (Nu f) where
-  readPrec = parens $ prec 10 $ do
-    Ident "fromFix" <- lexP
-    fromFix <$> step readPrec
-#endif
-
--- | A specialized, faster version of 'hoist' for 'Nu'.
-hoistNu :: (forall a. f a -> g a) -> Nu f -> Nu g
-hoistNu n (Nu next seed) = Nu (n . next) seed
-
-
+-- TODO: link from 'para' to 'zygo'
 zygo :: Recursive t => (Base t b -> b) -> (Base t (b, a) -> a) -> t -> a
 zygo f = gfold (distZygo f)
 
@@ -673,7 +700,99 @@ distGApoT
   -> f (ExceptT b m a)
 distGApoT g k = fmap ExceptT . k . fmap (distGApo g) . runExceptT
 
--- | Course-of-value iteration
+-- | A variant of 'cata' which includes the results of all the
+-- descendents, not just the direct children.
+--
+-- Like 'para', a sub-tree is provided for each recursive position. Each
+-- node in that sub-tree is annotated with the result for that
+-- descendent. The 'Cofree' type is used to add those annotations.
+--
+-- For our running example, let's recreate GitHub's directory compression
+-- algorithm. Notice that in [the repository for this
+-- package](https://github.com/recursion-schemes/recursion-schemes), GitHub
+-- displays @src\/Data\/Functor@, not @src@:
+--
+-- ![GitHub's code page](docs/github-compression.png)
+--
+-- GitHub does this because @src@ only contains one entry: @Data@. Similarly,
+-- @Data@ only contains one entry: @Functor@. @Functor@ contains several
+-- entries, so the compression stops there. This helps users get to the
+-- interesting folders more quickly.
+--
+-- Before we use 'histo', we need to define a helper function 'rollup'.
+-- It collects nodes until it reaches a node which doesn't have exactly one
+-- child. It also returns the labels of that node's children.
+--
+-- >>> :{
+-- let rollup :: [Cofree (TreeF node) label]
+--            -> ([node], [label])
+--     rollup [_ :< NodeF node cofrees] =
+--       let (nodes, label) = rollup cofrees
+--       in (node : nodes, label)
+--     rollup cofrees =
+--       ([], fmap extract cofrees)
+-- :}
+--
+-- >>> let foobar xs = 1 :< NodeF "foo" [2 :< NodeF "bar" xs]
+-- >>> rollup [foobar []]
+-- (["foo","bar"],[])
+-- >>> rollup [foobar [3 :< NodeF "baz" [], 4 :< NodeF "quux" []]]
+-- (["foo","bar"],[3,4])
+--
+-- The value @foobar []@ can be interpreted as the tree @NodeF "foo"
+-- [NodeF "bar" []]@, plus two annotations. The @"foo"@ node is annotated
+-- with @1@, while the @"bar"@ node is annotated with @2@. When we call
+-- 'histo' below, those annotations are recursive results of type @Int ->
+-- String@.
+--
+-- >>> :{
+-- let pprint5 :: Tree Int -> String
+--     pprint5 t = histo go t 0
+--       where
+--         go :: TreeF Int (Cofree (TreeF Int) (Int -> String))
+--            -> Int -> String
+--         go (NodeF node cofrees) indent
+--             -- cofrees :: [Cofree (TreeF Int) (Int -> String)]
+--             -- fs :: [Int -> String]
+--           = let indent' = indent + 2
+--                 (nodes, fs) = rollup cofrees
+--                 ss = map (\f -> f indent') fs
+--                 s = replicate indent ' '
+--                  ++ "* " ++ intercalate " / " (fmap show (node : nodes))
+--             in intercalate "\n" (s : ss)
+-- :}
+--
+-- >>> putStrLn $ pprint5 myTree
+-- * 0
+--   * 1
+--   * 2
+--   * 3 / 31 / 311
+--     * 3111
+--     * 3112
+--
+-- One common use for 'histo' is to cache the value computed for smaller
+-- sub-trees. In the Fibonacci example below, the recursive type is 'Natural',
+-- which is isomorphic to @[()]@. Our annotated sub-tree is thus isomorphic to
+-- a list of annotations. In our case, each annotation is the result which was
+-- computed for a smaller number. We thus have access to a list which caches
+-- all the Fibonacci numbers we have computed so far.
+--
+-- >>> :{
+-- let fib :: Natural -> Integer
+--     fib = histo go
+--       where
+--         go :: Maybe (Cofree Maybe Integer) -> Integer
+--         go Nothing = 1
+--         go (Just (_ :< Nothing)) = 1
+--         go (Just (fibNMinus1 :< Just (fibNMinus2 :< _)))
+--           = fibNMinus1 + fibNMinus2
+-- :}
+--
+-- >>> fmap fib [0..10]
+-- [1,1,2,3,5,8,13,21,34,55,89]
+--
+-- In general, @Cofree f a@ can be thought of as a cache that has the same
+-- shape as the recursive structure which was given as input.
 histo :: Recursive t => (Base t (Cofree (Base t) a) -> a) -> t -> a
 histo = gcata distHisto
 
@@ -698,11 +817,41 @@ gchrono w m = ghylo (distGHisto w) (distGFutu m)
 
 -- | Mendler-style iteration
 mcata :: (forall y. (y -> c) -> f y -> c) -> Fix f -> c
-mcata psi = psi (mcata psi) . unfix
+mcata psi = c where c = psi c . unFix
+
+-- | Mendler-style recursion
+--
+-- @since 5.2.2
+mpara :: (forall y. (y -> c) -> (y -> Fix f) -> f y -> c) -> Fix f -> c
+mpara psi = c where c = psi c id . unFix
+
+-- | Mendler-style semi-mutual recursion
+--
+-- @since 5.2.2
+mzygo :: (forall y. (y -> b) -> f y -> b) -> (forall y. (y -> c) -> (y -> b) -> f y -> c) -> Fix f -> c
+mzygo phi psi = c where c = psi c (mcata phi) . unFix
 
 -- | Mendler-style course-of-value iteration
 mhisto :: (forall y. (y -> c) -> (y -> f y) -> f y -> c) -> Fix f -> c
-mhisto psi = psi (mhisto psi) unfix . unfix
+mhisto psi = c where c = psi c unFix . unFix
+
+-- | Mendler-style coiteration
+--
+-- @since 5.2.2
+mana :: (forall y. (x -> y) -> x -> f y) -> x -> Fix f
+mana phi = c where c = Fix . phi c
+
+-- | Mendler-style corecursion
+--
+-- @since 5.2.2
+mapo :: (forall y. (Fix f -> y) -> (x -> y) -> x -> f y) -> x -> Fix f
+mapo phi = c where c = Fix . phi id c
+
+-- | Mendler-style course-of-values coiteration
+--
+-- @since 5.2.2
+mfutu :: (forall y. (f y -> y) -> (x -> y) -> x -> f y) -> x -> Fix f
+mfutu phi = c where c = Fix . phi Fix c
 
 -- | Elgot algebras
 elgot :: Functor f => (f a -> a) -> (b -> Either a (f b)) -> b -> a
@@ -728,15 +877,79 @@ zygoHistoPrepro f g t = gprepro (distZygoT f distHisto) g t
 -- Effectful combinators
 -------------------------------------------------------------------------------
 
--- | Effectful 'fold'.
+-- | A specialization of 'cata' for effectful folds.
 --
--- This is a type specialisation of 'cata'.
+-- 'cataA' is the same as 'cata', but with a more specialized type. The only
+-- reason it exists is to make it easier to discover how to use this library
+-- with effects.
 --
--- An example terminating a recursion immediately:
+-- For our running example, let's improve the output format of our
+-- pretty-printer by using indentation. To do so, we will need to keep track of
+-- the current indentation level. We will do so using a @Reader Int@ effect.
+-- Our recursive positions will thus contain @Reader Int String@ actions, not
+-- @String@s. This means we need to run those actions in order to get the
+-- results.
 --
--- >>> cataA (\alg -> case alg of { Nil -> pure (); Cons a _ -> Const [a] })  "hello"
--- Const "h"
+-- >>> :{
+-- let pprint2 :: Tree Int -> String
+--     pprint2 = flip runReader 0 . cataA go
+--       where
+--         go :: TreeF Int (Reader Int String)
+--            -> Reader Int String
+--         go (NodeF i rss) = do
+--           -- rss :: [Reader Int String]
+--           -- ss  :: [String]
+--           ss <- local (+ 2) $ sequence rss
+--           indent <- ask
+--           let s = replicate indent ' ' ++ "* " ++ show i
+--           pure $ intercalate "\n" (s : ss)
+-- :}
 --
+-- >>> putStrLn $ pprint2 myTree
+-- * 0
+--   * 1
+--   * 2
+--   * 3
+--     * 31
+--       * 311
+--         * 3111
+--         * 3112
+--
+-- The fact that the recursive positions contain 'Reader' actions instead of
+-- 'String's gives us some flexibility. Here, we are able to increase the
+-- indentation by running those actions inside a 'local' block. More generally,
+-- we can control the order of their side-effects, interleave them with other
+-- effects, etc.
+--
+-- A similar technique is to specialize 'cata' so that the result is a
+-- function. This makes it possible for data to flow down in addition to up.
+-- In this modified version of our running example, the indentation level flows
+-- down from the root to the leaves, while the resulting strings flow up from
+-- the leaves to the root.
+--
+-- >>> :{
+-- let pprint3 :: Tree Int -> String
+--     pprint3 t = cataA go t 0
+--       where
+--         go :: TreeF Int (Int -> String)
+--            -> Int -> String
+--         go (NodeF i fs) indent
+--             -- fs :: [Int -> String]
+--           = let indent' = indent + 2
+--                 ss = map (\f -> f indent') fs
+--                 s = replicate indent ' ' ++ "* " ++ show i
+--             in intercalate "\n" (s : ss)
+-- :}
+--
+-- >>> putStrLn $ pprint3 myTree
+-- * 0
+--   * 1
+--   * 2
+--   * 3
+--     * 31
+--       * 311
+--         * 3111
+--         * 3112
 cataA :: (Recursive t) => (Base t (f a) -> f a) -> t -> f a
 cataA = cata
 
@@ -797,8 +1010,9 @@ transverse n = cata (fmap embed . n)
 -- >>> data Pair a = Pair a a deriving Functor
 --
 -- >>> :{
--- let zipWith' :: (a -> a -> b) -> [a] -> [a] -> [b]
+-- let zipWith' :: forall a b. (a -> a -> b) -> [a] -> [a] -> [b]
 --     zipWith' f xs ys = cotransverse g (Pair xs ys) where
+--       g :: Pair (ListF a c) -> ListF b (Pair c)
 --       g (Pair Nil        _)          = Nil
 --       g (Pair _          Nil)        = Nil
 --       g (Pair (Cons x a) (Cons y b)) = Cons (f x y) (Pair a b)
